/* Downloader for MACSBUG.  
   Author Anonymous.
   Modified to store symbols: V. Pratt, Jan. 1981.
   Assumed to run on a machine with VAX-type byte order in a word,
   namely first text byte is least significant byte.  Modify the function
   reverse(lwrd) appropriately to run on another machine.
   Jan. 1982 - V. Pratt - Modified to load symbol table where the Sun
   	bootloader puts it
   12 July 1985	Jeff Mogul	Stanford
   	Enlarged symbol table storage space
 */

#include <stdio.h>
#include <a.out.h>

#define STARTADDR 0x1000	/* Default start address */
#define RCDSIZE 12		/* Size of Motorola S-type records */
#define MACSSYMTAB 0x6BA	/* Symbol table starts here - MACSBUG */
#define SUNSYMTAB  0x1f000	/* Symbol table starts here - SUN */
#define SYMTAB (vflag=='m'?MACSSYMTAB:SUNSYMTAB)
#define SYMTABADDR 0x570	/* Boundaries of symbol table here */
#define MAXSYMS (vflag=='m'?((0x1000-0x6BA)/12):(0x1000/12))
#define MAXSYMST (0x1000/12)	/* Max size of symbol table for either
					SUN or MACSBUG */

#define hex(c) c>9 ? c+0X37 : c+0X30  /* macro for hex convert */

struct exec filhdr;

struct sym68 {			/* fake sym for 68000 */
	short dummy1;		/* ignored */
	unsigned char stype;	/* type */
	unsigned char slength;	/* length */
	long svalue;
};

char	Sdata=2;    	    	/* Default to S2/S8 records (24 bit addr) */
char	Send=8;
char	vflag=0;		/* 1 -> Design Module */
FILE *infile, *outfile;
char infilename[30], outfilename[30] = {0};
int syms = 0;
char stable[MAXSYMST*(sizeof(struct sym68) + 80)];	/* enough? */
int evenprom = 0;
int oddprom = 0;
int textout = 1;
int dataout = 1;

puthex(b)			/* print byte as two hex chars */
char b;
{
register char c1,c2;

    c1=(b>>4) & 0XF; c2=b & 0XF;
    c1=hex(c1); c2=hex(c2);
    fprintf(outfile,"%c%c",c1,c2);
}

typedef struct Msymtab 		/* structure of the MACSBUG symbol space */
    {char name[8];
     int value;
    } msymtab;

#ifdef RBO
reverse(lwrd) unsigned lwrd;
 {return((lwrd>>24)	     |
	 (lwrd>>8 & 0xff00)  |
	 (lwrd<<8 & 0xff0000)|
	 (lwrd<<24)
	);
 }
#else
#define reverse(lwrd) (lwrd)
#endif

int checksum;

checkout(c) char c; {checksum += c; puthex(c);}

/* Procedure to print out one ExorMacs record of given type */

print_exormacs_record(type,buffer,Dcount,addr)
    char *buffer; int Dcount, addr;
{
    int reccount, recaddr;

    if (evenprom || oddprom) {
	reccount = (Dcount+1)/2;		/* half number of bytes */
	recaddr = (addr+1)/2;
    }
    else {
	reccount = Dcount;			/* full record */
	recaddr = addr;
    }
    fprintf(outfile,"S%d",type);
    checksum = 0;
    if (type == 2 || type == 8) {
    	checkout((char) reccount+4);
    	checkout((char) (recaddr>>16));
    }
    else
 	checkout((char) reccount+3);
    checkout((char) (recaddr>>8));
    checkout((char) recaddr);
    while (Dcount--) {
	if (!evenprom && !oddprom)
	    checkout(*buffer);
	else if (evenprom && !(Dcount & 1))
	    checkout(*buffer);
	else if (oddprom && (Dcount & 1))
	    checkout(*buffer);
	buffer++;	
    }
    puthex(~(checksum & 0XFF));
    fprintf(outfile,"\n");
}

/* Procedure that puts out records from a buffer */
outbuf(buffer,addr,size)
char *buffer;
int addr, size;
{
register int count,		/* counts total number of bytes processed*/
	     Dcount;		/* number of bytes in current record */

    for(count=0; count<size; count += RCDSIZE) 
	{Dcount= (size-count<RCDSIZE) ? size-count : RCDSIZE;
	 print_exormacs_record(Sdata,buffer,Dcount,addr);
	 buffer += Dcount;
	 addr += Dcount;
	}
}

/* Procedure that puts out records  from a file */
outbin(file,addr,size)
FILE *file;
int addr, size;
{
register int count,		/* counts total number of bytes processed*/
	     Dcount;		/* number of bytes in current record */
char buffer[RCDSIZE];		/* buffer for data */

    for(count=0; count<size; count += RCDSIZE) 
	{Dcount= (size-count<RCDSIZE) ? size-count : RCDSIZE;
	 if(fread(buffer,Dcount,1,file) != 1)
	     {fprintf(stderr,"Read error\n"); exit(1);}
	 print_exormacs_record(Sdata,buffer,Dcount,addr);
	 addr += Dcount;
	}
}

#define lower(c) 'a' <= c && c <= 'z'
#define upper(c) 'A' <= c && c <= 'Z'
#define digit(c) '0' <= c && c <= '9'


main(argc,argv)
int argc;
char *argv[];
{
int Sa, Daddr=STARTADDR, i;
msymtab table[MAXSYMST];

    if(argc<2)
	{fprintf(stderr,"Usage: dl68 -tdEOTvos infile [ > outfile ]\n");
	 exit(1);
	}
    for (i=1; i<argc; i++)				/* get options */
	if (argv[i][0] == '-')
	switch (argv[i][1]) {
		case 't':		/* t - text only */
			textout = 1;
			dataout = 0;
			break;
		case 'd':		/* d - data only */
			dataout = 1;
			textout = 0;
			break;
		case 'E':
			evenprom = 1;
			break;
		case 'O':
			oddprom = 1;
			break;
		case 'T': if (i++ < argc)
			    sscanf(argv[i],"%x",&Daddr); 
			    break;
		case 'v':
			vflag = argv[i][2];
			break;
		case 'o':
			if (i++ < argc)
				strcpy(outfilename,argv[i]);
			break;
    	    	case 's':
    	    	    	Sdata = argv[i][2] - '0';
    	    	    	Send = argv[i][3] - '0';
    	    	    	break;
	}
	else strcpy(infilename,argv[i]);
    if((infile=fopen(infilename,"r"))==NULL)	 /* open infile */
       {strcat(infilename,".68");     			 /* try .68 form */
	if ((infile=fopen(infilename,"r"))==NULL) 
           {fprintf(stderr,"dl68: Can't open %s\n",infilename);
	    exit(1);
	   }
       }
    if (*outfilename) {
	if ((outfile = fopen(outfilename,"w")) == NULL) {
		fprintf(stderr,"dl68: Can't open %s\n",outfilename);
		exit(1);
	}
    }
    else outfile = stdout;
    if(fread(&filhdr, sizeof(struct exec),1,infile) != 1)/* get header info */
	{fprintf(stderr,"dl68: %s wrong format\n",infilename);
	 exit(1);
	}
    if(N_BADMAG(filhdr)) {    /* check magic number */
	 fprintf(stderr,"dl68: %s not proper a.out file\n",infilename); 
	 exit(1);
    }

    Sa = Daddr;

    /*
     * Output records for TEXT portion
     */
    if (textout) {
	fseek(infile,N_TXTOFF(filhdr),0);	    /* seek to DATA portion */
	outbin(infile,Daddr,filhdr.a_text);
	Daddr += filhdr.a_text;    /* adjust output address */
    }

    /*
     * Output records for DATA segment
     */
    if (dataout) {
	fseek(infile,N_DATOFF(filhdr),0);	    /* seek to DATA portion */
	outbin(infile,Daddr,filhdr.a_data); /* write data */
	Daddr += filhdr.a_data;		    /* adjust output address */
    }

    /*
     * Output entry point: as in file header, else default to start address
     */
    print_exormacs_record(Send,0,0,filhdr.a_entry? filhdr.a_entry: Sa);
    exit(0);
}

