/*
 * $Id: atalkad.c,v 1.20 89/09/04 00:09:20 djh Exp Locker: djh $
 */

#ifndef lint
static char rcsid[]  = "@(#)$Id: atalkad.c,v 1.20 89/09/04 00:09:20 djh Exp Locker: djh $";
static char sccsid[] = "@(#)atalkad.c	1.2 (Stanford) 9/87";
#endif

/*
 * Appletalk administration daemon.
 *
 * Answers request packets from client appletalk gateways by supplying
 * the appletalk-to-ip network configuration table and the local
 * appletalk gateway configuration information.
 */

/*
 * (C) 1986, Stanford Univ.  CSLI.
 * May be used but not sold without permission.
 */

/*
 * $Log:	atalkad.c,v $
 * Revision 1.20  89/09/04  00:09:20  djh
 * Add code to allow many more than 64 appletalk nets (an early KIP limitation)
 * Do this by designing a protocol to allow a gateway to know and
 * be able to ask for more than one aaROUTEI or aaPROXY packet (kre & djh)
 * 
 * Revision 1.15  89/04/07  02:05:35  kre
 * Add 'M' line type for multigates, to squeeze a few net numbers out of
 * the routing tables so we can add more before we run out of space.
 * Sort and compress the proxy table, uses much less space and is much
 * neater.  Alloy 0x for hex humbers (also 0h!) as well as the bare x or
 * h, and also leading 0 for octal (this may break atalkatab files).
 * Get rid of the _ -> space abomination and support quoting instead.
 * Don't add a trailing null to pascal type strings ('...') in the config
 * data section, this type was never documented, and is never used, but
 * should be done correctly.  Several straight out bug fixes.  Lots of
 * fluff removed (almost lint free now, just difference of opinion about
 * what type signal functions should be, void or int).  Also added VARARGS
 * support (defined with -DVARARGS) for hosts that need varargs to work.
 * 
 * Revision 1.14  89/03/21  00:29:08  kre
 * Upgrade for KIP0688 compatability - needed for Megan 1.14 ->
 * 
 * Revision 1.10  89/03/20  23:34:48  kre
 * Initial version (Megan 1.10->1.13)
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <netinet/in.h>
#define iaddr_t unsigned long
#include "gwctl.h"
#include "atalk.h"

#include <signal.h>
#include <stdio.h>
#include <strings.h>
#include <errno.h>
#include <ctype.h>
#include <netdb.h>

/* for 4.2 systems */
#ifndef FD_SETSIZE
# define FD_SETSIZE sizeof(int)*8
# define NFDBITS sizeof(int)*8
# define howmany(x,y) (1)
# define FD_SET(n, p)  (p)->fds_bits[0] |= (1<<(n))
# define FD_CLR(n, p)  (p)->fds_bits[0] &= ~(1<<(n))
# define FD_ISSET(n, p) ((p)->fds_bits[0] & (1<<(n)))
# define FD_ZERO(p) bzero((char *)p, sizeof(*p))
#endif

#ifndef ATALKATAB
#define	ATALKATAB "/etc/atalkatab"
#endif
#ifndef ATALKALOG
#define	ATALKALOG "/usr/adm/atalkalog"
#endif
#ifndef ATALKAPID
#define	ATALKAPID "/etc/atalkapid"
#endif

int	debug;
extern	int errno;
struct	sockaddr_in sin = { AF_INET };
int	s;		/* socket fd */
struct	sockaddr_in fsin; /* foreign sockaddr_in */
int	fsinlen;
int	sig, sigint(), sighup();
struct aaconf aa;	/* receive packet */
struct aaconf aas;	/* send packet */
struct timeval tvwait;	/* timer for select */

/* Mask of high 3 bytes of a network address */
iaddr_t node_mask;
  
/*
 * Globals below are associated with the atalka database file (atalkatab).
 */
char	*atalkatab = ATALKATAB;
char	*atalkalog = ATALKALOG;
char	*atalkapid = ATALKAPID;
FILE	*fp;
char	line[256];	/* line buffer for reading atalkatab */
char	*linep;		/* pointer to 'line' */
int	linenum;	/* current line number in atalkatab */

#define ROUTEPKTSIZ 512	/* the absolute maxm */
#define	NANETS	1024	/* max number of 'anets' structs */

struct anets {
	u_short	net;		/* atalk net */
	u_char	flags;		/* flags, see aroute* in gwctl.h */
	u_char	confsize;	/* size of databytes in conf below */
	iaddr_t	iaddr;		/* ip address */
	char	zone[32];	/* zone name */
	char	conf[64];	/* configuration info, if kbox */
} anets[NANETS];

#define	typeFlags	(arouteKbox|arouteNet|arouteHost|arouteBMask)

#define	NMGATES	128	/* max number of multigates */

struct mgates {
	struct	anets	*first;
	struct	anets	*last;
} mgates[NMGATES];

/* Internal format of conf structure */
struct conf {
	iaddr_t	ipbroad;		/* broadcast addr on ether */
	iaddr_t	ipname;			/* address of name server */
	iaddr_t	ipdebug;		/* address of debug host */
	iaddr_t	ipfile;			/* address of file server */
	u_long	ipother[4];		/* other addresses passed via IPGP */
#define ipsmask anetet		        /* old use for this field */
	u_short	anetet;		        /* ethertalk net number of enet */
#define ipsshift startddpWKSUnix	/* old: subnet shift (unused) */
	u_short	startddpWKSUnix;	/* start of unix WKS udp ports */
	u_long	flags;			/* various bit flags */
#define	conf_stayinzone 0x1		/* no looking at other zones */
#define conf_laserfilter 0x2		/* NBP filtering for LaserWriters */
#define	conf_tildefilter 0x4		/* NBP filtering, "name~" */
	u_short	ipstatic;		/* number of static IP addrs */
	u_short	ipdynamic;		/* number of dynamic IP addrs */
	u_short	atneta;			/* atalk net #, appletalk */
	u_short	atnete;			/* atalk net #, ethernet */
} conf_proto;

int	nanets;		/* current number of anets */
int	nmgates;	/* number of multigtates */
long	modtime;	/* last modification time of atalkatab */
long	size;		/* last known size of file */
char	aroutes[4096];	/* route tuples built by buildart() */
int	arouteslen;
char	azones[2048];	/* zone table built by buildzone() */
int	azoneslen;
char	aproxy[8192];	/* nbp proxy table built by buildproxy() (> 128*5*12) */
int	aproxylen;

int scanonly = 0;

#define	Printf	(void)printf		/* de lint */
#define	Fprintf	(void)fprintf		/* de lint */

usage()
{
  Printf("usage: atalkad [-c filename] [-debug] [route|boot|exit]\n");
  Printf("usage:  -c means check file\n");
  exit(1);
}

main(argc, argv)
	char *argv[];
{
	register int n;
	FILE *filep;
	int pid;
	unsigned long inet_addr();

	for (argc--, argv++ ; argc > 0 ; argc--, argv++) {
		if (argv[0][0] == '-') {
			switch (argv[0][1]) {
			case 'd':
				debug++;
				break;
			case 'c':
				if (argv[0][2]) {
				  atalkatab = argv[0]+2;
				} else if (argc > 1 && argv[1]) {
				  atalkatab = argv[1];
				  argc--, argv++;
				}
				scanonly++;
				break;
			default:
				usage();
			}
			continue;
		}
		if (scanonly)
			usage();
		if (strcmp(argv[0], "boot") == 0)
			sig = SIGINT;
		else if (strcmp(argv[0], "route") == 0)
			sig = SIGHUP;
		else if (strcmp(argv[0], "exit") == 0)
			sig = SIGKILL;
		else
			usage();
		if ((filep = fopen(atalkapid, "r")) == NULL
		    || fscanf(filep, "%d", &pid) != 1
		    || kill(pid, sig) < 0) {
			Printf("failed to send signal to daemon\n");
			exit(1);
		} else {
			Printf("sent signal to daemon\n");
			exit(0);
		}
	}
	
	if (scanonly) {
		readtab();
		dumptab();
		exit(0);
	}

	if (debug == 0) {
		register int f;
		register char *p;
		extern char *rindex();

		if (fork())
			exit(0);
		for (f = getdtablesize(); --f >= 0; )
			(void) close(f);
		(void) open("/", 0);
		(void) dup2(0, 1);
		(void) dup2(0, 2);
#ifdef TIOCNOTTY
		f = open("/dev/tty", 2);	
		if (f >= 0) {
			(void)ioctl(f, TIOCNOTTY, (char *)0);
			(void) close(f);
		}
#endif
		p = rindex(atalkatab, '/');
		if (p)
			*p = 0;
		(void)chdir(*atalkatab ? atalkatab : "/");
		if (p)
			*p = '/';
	}

	(void)signal(SIGHUP, sighup);
	(void)signal(SIGINT, sigint);
	pid = getpid();

	while ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		log("ATALKAD(%d): socket call failed", pid);
		sleep(30);
	}
	sin.sin_port = htons(aaPort);
	if (bind(s, (struct sockaddr *)&sin, sizeof (sin)) < 0) {
		log("ATALKAD(%d): bind call failed (another daemon?)", pid);
		exit(1);
	}

	log("###	ATALKA daemon starting (%d)", pid);
	if ((filep = fopen(atalkapid, "w")) == NULL) {
		log("couldnt create pid file %s\n", atalkapid);
		exit(1);
	}
	Fprintf(filep, "%d\n", pid); 
	(void)fclose(filep);

	/* fill in node mask in a machine independent manner */
	node_mask = inet_addr("255.255.255.0");

	readtab();
	for (;;) {
		fsinlen = sizeof (fsin);
		n = recvfrom(s, (caddr_t)&aa, sizeof aa, 0,
		    (struct sockaddr *)&fsin, &fsinlen);
		if (n < 0) {
			if (errno != EINTR) {
				log("recv failed");
				exit(1);
			}
			readtab();
			if (sig != 0)
				sendall(sig);
			sig = 0;
			continue;
		}
		if (n < aaconfMinSize || ntohl(aa.magic) != aaMagic)
			continue;
		readtab();
		sendreply();
	}
}


/*
 * Interrupts are used to signal type of sendall().
 */
sighup() { sig = aaROUTEI; }
sigint() { sig = aaRESTART; }


/*
 * Send reply to packet aa.
 */
sendreply()
{
	register struct anets *an;
	register n;
	register offset;
	struct in_addr A;
	char *inet_ntoa();

	A.s_addr = aa.ipaddr;
	switch (aa.type) {
	case aaCONF:
		for (an = &anets[0], n = 0 ; n < nanets ; an++, n++)
			if (an->iaddr == aa.ipaddr && an->confsize &&
			    (an->flags & arouteKbox) &&
			    (an->flags & arouteEtalk) == 0)
				goto found;
		log("aaCONF from %s ***** not in table", inet_ntoa(A));
		return;
found:
		log("aaCONF to %s", inet_ntoa(A));
		n = an->confsize;
		bcopy(an->conf, (char *)aa.stuff, n);
		break;
	case aaZONE:
		log("aaZONE to %s", inet_ntoa(A));
		n = azoneslen;
		bcopy(azones, (char *)aa.stuff, n);
		break;
	case aaZONEQ:
		log("aaZONEq from %s", inet_ntoa(A));
		if ((n = zipinput()) <= 0)
		  return;	/* drop */
		log("aaZONEq reply to %s (len %d)", inet_ntoa(A), n);
		break;
	case aaPROXYQ:
	case old_aaPROXYQ:		/*ZZZ* transition stuff */
		log("aaPROXY to %s", inet_ntoa(A));
		if(aa.flags == 0)
			offset = 0;
		else {
			offset = (aa.flags-1)*504;
			if(offset > aproxylen) { /* trouble */
				offset = 0;
				log("Bozo aaPROXYQ flag length");
			}
		}
		n = aproxylen - offset;
		if(n > 504)
			n = 504;
		bcopy(aproxy+offset, (char *)aa.stuff, n);
	/*ZZ				  ZZZ  transition stuff
		aa.type = aaPROXY;
	*ZZ*/	aa.type--;		/*ZZZ* transition stuff */
		aa.flags = (aproxylen / 504) + 1; /* 504/12 = 42 */
		break;
	case aaROUTEI:
	case aaROUTER:		/* extend the routing info a bit */
		if(aa.type == aaROUTEI) {
			log("aaROUTEI to %s", inet_ntoa(A));
			offset = 0;
			if(arouteslen > ROUTEPKTSIZ)
				aa.flags=(arouteslen+ROUTEPKTSIZ-1)/ROUTEPKTSIZ;
			else
				aa.flags = 0;
		} else {
			log("aaROUTER from %s", inet_ntoa(A));
			if((offset = aa.flags*ROUTEPKTSIZ) > arouteslen) {
				offset = arouteslen;
				log("Bozo aaROUTER request");
			}
			log("aaROUTEM to %s", inet_ntoa(A));
			aa.type = aaROUTEM;
			/* and don't touch the flags */
		}
		n = arouteslen - offset;
		if(n > ROUTEPKTSIZ)
			n = ROUTEPKTSIZ;
		bcopy(aroutes+offset, (char *)aa.stuff, n);
		break;
	default:
		/* don't reply here! */
		return;
	}
	aa.count = htons((unsigned short)n);
	if (sendto(s, (caddr_t)&aa, aaconfMinSize+n, 0,
	    (struct sockaddr *)&fsin, sizeof fsin) < 0)
		log("Sendto to %s failed", inet_ntoa(fsin.sin_addr));
}


char *
netstring(n)
u_short n;
{
  int net;
  static char mynetstr[30];
  extern char *sprintf();

  net = ntohs(n);
  (void)sprintf(mynetstr, "%d.%d", (net>>8) & 0xff, net & 0xff);
  return(mynetstr);
}

/*
 * return pointer to zone name
 * NULL if none
*/
char *
zoneof(net)
u_short net;
{
  register struct anets *an;
  int n;

  for (an = &anets[0], n=0 ; n < nanets ; an++, n++)
    if (net == an->net)
      return(an->zone);
  return(NULL);
}

/*
 * return a value that is always printable
 *
*/
char *
pzoneof(net)
u_short net;
{
  char *p = zoneof(net);
  return(p == NULL ? "unknown" : p);
}

dumptab()
{
  register struct anets *an;
  register u_long *lp;
  register n;
  register struct mgates *mg, *lmg;
  char *hostnameof();
  struct conf *conf;
  iaddr_t rs, re;
  struct in_addr A;

  size = 0;
  mg = mgates;
  lmg = &mgates[nmgates];
  for (an = &anets[0], n = 0 ; n < nanets ; an++, n++) {
    while (mg < lmg && mg->last < an)
	mg++;
    Printf("Route %s %s,\t", netstring(an->net), an->zone);
    if ((an->flags & arouteKbox) && (an->flags & arouteEtalk) == 0) {
      if (mg < lmg && an >= mg->first && an <= mg->last)
	if (an == mg->first)
	  Printf("Multigate");
	else
	  Printf("Multigate iface %d", an - mg->first - 1);
      else
	Printf("Kinetics box");
    }
    if (an->flags & arouteCore)
      Printf(" core gateway");
    if (an->flags & arouteHost) {
      Printf("Host is redirector");
      if (an->flags & arouteAsync)
	Printf(" to an async network");
    }
    if (an->flags & arouteNet)
	Printf(" net %d", an->flags & arouteBMask);
    if ((an->flags & (arouteKbox|arouteEtalk)) == (arouteKbox|arouteEtalk))
      Printf("EtherTalk");
    if ((an->flags & typeFlags) == arouteKbox &&
	mg < lmg && an > mg->first && an <= mg->last) {
      A.s_addr = an->iaddr;
      Printf(" at %s", hostnameof(mg->first->iaddr));
      Printf(" interface %s", inet_ntoa(A));
    } else
      Printf(" at %s", hostnameof(an->iaddr));
    (void)putchar('\n');
    if ((an->flags & arouteKbox) && (an->flags & arouteEtalk) == 0) {
      conf = (struct conf *)an->conf;
      A.s_addr = conf->ipbroad;
      Printf("\tIP Broadcast: %s, ", inet_ntoa(A));
      if (conf->ipname == conf->ipdebug) {
	Printf("IP name and debug: %s\n", hostnameof(conf->ipname));
	Printf("\tIP file server (unused)", hostnameof(conf->ipfile));
      } else {
	Printf("IP name: %s\n",hostnameof(conf->ipname));
	Printf("\tIP debug %s, ", hostnameof(conf->ipdebug));
	Printf("IP file server (unused)", hostnameof(conf->ipfile));
      }
      if (ntohl(conf->flags) & conf_stayinzone)
	Printf(", marked stay in zone");
      if (ntohl(conf->flags) & conf_laserfilter)
	Printf(", laserwriters stay in zone");
      if (ntohl(conf->flags) & conf_tildefilter)
	Printf(", tilde marked names stay in zone");
      (void)putchar('\n');

      if (conf->startddpWKSUnix == htons(defddpWKSUnix))
	Printf("\tUDP Port range for WKS starts at %d (old range)\n",
	       defddpWKSUnix);
      else
	Printf("\tUDP Port range for WKS starts at %d\n",
	       ntohs(conf->startddpWKSUnix));

#ifdef SHORTFORMAT
      if (conf->ipstatic)
	Printf("\t%d ", ntohs(conf->ipstatic));
      else
	Printf("\tNo");
      Printf(" static ip addresses, ");
      if (conf->ipdynamic)
	Printf("\t%d ", ntohs(conf->ipdynamic));
      else
	Printf("\tNo");
      Printf(" dynamic ip addresses\n");
#else
      re = rs = ntohl(an->iaddr);
      if (conf->ipstatic) {
	rs++;
	A.s_addr = htonl(rs);
	Printf("\tIP static address range: %s", inet_ntoa(A));
	re += ntohs(conf->ipstatic);
	A.s_addr = htonl(re);
	Printf(" %s\n", inet_ntoa(A));
      } else
	Printf("\tIP static address range: empty\n");

      if (conf->ipdynamic) {
	rs = re + 1;
	re += ntohs(conf->ipdynamic);
	A.s_addr = htonl(rs);
	Printf("\tIP dynamic address range: %s", inet_ntoa(A));
	A.s_addr = htonl(re);
	Printf(" %s\n", inet_ntoa(A));
      } else
	Printf("\tIP dynamic address range: empty\n");
#endif

      Printf("\tKbox interfaces:\n");
      Printf("\t\tlocaltalk: %s zone %s\n", netstring(conf->atneta),
	     pzoneof(conf->atneta));
      Printf("\t\tKIP: %s zone %s\n", netstring(conf->atnete),
	     pzoneof(conf->atnete));
      if (conf->anetet) {
	Printf("\t\tEtherTalk: %s zone %s\n", netstring(conf->anetet),
	       pzoneof(conf->anetet));
      }
    }
  }
  Printf("\nNBP proxy table:\n");
  for (lp = (u_long *)aproxy; lp < (u_long *)&aproxy[aproxylen]; ) {
    rs = (iaddr_t) *lp++;
    if (rs == 0)
      break;
    re = (iaddr_t) *lp++;
    n = *lp++;
    if (n == 0)
      Printf("\tAppletalk: ");
    else
      Printf("\tExternal:  ");
    A.s_addr = rs;
    Printf("%s .. ", inet_ntoa(A));
    A.s_addr = re;
    Printf("%s\n", inet_ntoa(A));
  }
}

/*
 * Build arouteTuple's into area provided by caller.
 * Return byte count of tuples deposited.
 */
buildart(at, maxsize)
	register struct arouteTuple *at;
        int maxsize;
{
	register struct anets *an;
	register struct mgates *mg, *lmg;
	register n, size;

	size = 0;
	mg = mgates;
	lmg = &mgates[nmgates];
	for (an = &anets[0], n = 0 ; n < nanets ; an++, n++) {
		while (mg < lmg && mg->last < an)
			mg++;
		if ((an->flags & typeFlags) == arouteKbox &&
		    mg < lmg && an == mg->first)
			continue;

		size += sizeof *at;
		if (size > maxsize) {
			size -= sizeof *at;
			log("art build: routing table too big!!");
			break;
		}

		if ((an->flags & typeFlags) == arouteKbox &&
		    mg < lmg && an > mg->first && an <= mg->last)
			at->node = mg->first->iaddr;
		else
			at->node = an->iaddr;

		at->net = an->net;
		at->flags = an->flags;
#ifdef KIP0688
		at->hops = 0; /* start at one - incremented in gateway */
#else
		at->hops = 1; /* not 0, so route appears to be 2 hops in gw */
			   /* this causes gateway to prefer ethertalk route */
#endif
		at++;
	}
	log("art build: %d entries, %d maximum",
	    size/(sizeof(struct arouteTuple)),
	    maxsize/(sizeof(struct arouteTuple)));
	if (size > (maxsize - 40))
		log("art build: routing table size %d is near maximum of %d",
		    size, maxsize);
	return (size);
}


#define	MAXZ	32

/*
 * Build zones structure to be returned by aaZONE.
 */
buildzones(az, azlen)
	char *az;
	int azlen;
{
	char zname[MAXZ][33];
	int nzname = 0;
	char netzone[NANETS];
	register i, iz, n;
	register char *cp;
	int zonenamelen = 0;

	/* make a pass thru anets, finding all unique zone names. */

	for (i = 0 ; i < nanets ; i++) {
		if (anets[i].zone[0] == 0) {
			netzone[i] = MAXZ+1;
			continue;
		}
		for (iz = 0 ; iz < nzname ; iz++)
			if (strcmp(anets[i].zone, zname[iz]) == 0)
				goto found;

		n = strlen(anets[i].zone) + 1;
		if (nzname >= MAXZ || zonenamelen + n >= azlen) {
			log("zone name table overflow");
    log("ignore this if running KIP 06/88, Megan 1.14, or later versions");
			netzone[i] = MAXZ+1;
			continue;
		}
		/* not found, make a new name */
		(void)strcpy(zname[iz], anets[i].zone);
		zonenamelen += n;
		nzname++;

	found:;	/* found or inserted it, note the zname index */
		netzone[i] = iz;
	}
	/*
	 * Build structure to send to gateway.  Looks like:
	 * net# net# ... 0 zonename
	 * net# net# ... 0 zonename
	 * 0xFFFF.
	 */
	cp = az;
	for (iz = 0 ; iz < nzname ; iz++) {
		for (i = 0 ; i < nanets ; i++) {
			if (netzone[i] != iz)
				continue;
			/* anets[].net has been swapped by readtab */
			*cp++ = (ntohs(anets[i].net) >> 8);
			*cp++ = ntohs(anets[i].net);
		}
		*cp++ = 0;  *cp++ = 0;
		*cp++ = n = strlen(zname[iz]);
		bcopy(zname[iz], cp, n);
		cp += n;
	}
	*cp++ = 0xFF;  *cp++ = 0xFF;
	n = cp - az;		/* length of table */
	if (n > azlen) {
	  log("buildzones: ZIP table too large");
   log("Don't worry if you are using KIP 06/88, Megan 1.14, or later versions");
	} else if (n > (azlen-30))
	  log("buildzones: ZIP table size %d approaching maximum of %d",
	      n, azlen);
	log("zone names take %d bytes in gateway", n);
	return (n);
}

/*
 * Build NBP proxt table, to be returned by aaPROXYQ
 */
buildproxy(pp, nl)
	long *pp;
	register nl;
{
	register long *lp = pp;
	register struct anets *ap;
	register i = nanets;
	register long *cp, *np, *op;
	int proxcmp();
	extern char *malloc();

#define	Conf(p)	((struct conf *)(p))

	ap = anets;
	while (nl > 2) {	/* each entry requires 3 longs */
		if (--i < 0)
			break;
		*lp++ = ap->iaddr;
		if (ap->flags & arouteKbox) {
			if (ap->confsize >= sizeof(u_short) +
			  (char *)(&Conf(ap->conf)->ipdynamic)-(char *)ap->conf)
				*lp++ = htonl(ntohl(ap->iaddr) +
				    ntohs(Conf(ap->conf)->ipstatic) +
				    ntohs(Conf(ap->conf)->ipdynamic));
			else
				*lp++ = ap->iaddr;
			*lp++ = htonl((u_long)0);
		} else if (ap->flags & arouteHost) {
			*lp++ = ap->iaddr;
			*lp++ = htonl((u_long)1);
		} else if (ap->flags & arouteNet) {
			/*
			 * this fails completely for N0, but, we can't have
			 * everything (the info just isn't available)
			 */
			*lp++ = htonl(ntohl(ap->iaddr) +
			    (1 << (ap->flags & arouteBMask)*8));
			*lp++ = htonl((u_long)1);
		}
		ap++;
		nl -= 3;
	}
	if (nl > 0) {
		*lp++ = htonl((u_long)0);
		if (nl < 10)
			log("Warning: approaching proxy table overflow");
	} else
		log("Proxy table overflow");

	i = (caddr_t)lp - (caddr_t)pp;
	if (i == 0)
		return (0);

	qsort((char *)pp, i / (3 * sizeof(long)), 3 * sizeof(long), proxcmp);

	op = lp;
	np = (long *)malloc((unsigned)i);

	lp = pp;
	cp = np;

	*cp++ = *lp++;
	*cp++ = *lp++;
	*cp++ = *lp++;

	while (lp < op) {
		if (lp[0] == lp[-3] && lp[1] == lp[-2]) {
			cp[-1] = lp[2];
			lp += 3;
			continue;
		}
		if (ntohl((u_long)lp[-2]) + 1 == ntohl((u_long)lp[0]) &&
		    lp[-1] == lp[2]) {
			cp[-2] = lp[1];
			lp += 3;
			continue;
		}
		if (ntohl((u_long)lp[-2]) + 3 == ntohl((u_long)lp[0]) &&
		    lp[-1] == lp[2]) {
			i = ntohl((u_long)lp[-2]) ^ ntohl((u_long)lp[0]);
			if (i ^ (i + 1) == 0) {
				cp[-2] = lp[1];
				lp += 3;
				continue;
			}
		}
		*cp++ = *lp++;
		*cp++ = *lp++;
		*cp++ = *lp++;
	}

	i = (caddr_t)cp - (caddr_t)np;
	bcopy((char *)np, (char *)pp, i);
	free((char *)np);

	log("Proxy table uses %d bytes", i);
	return (i);
}

proxcmp(a, b)
	char *a, *b;
{
	register long *al, *bl;

	al = (long *)a;
	bl = (long *)b;

	if (al[0] < bl[0])
		return (-1);
	if (al[0] > bl[0])
		return (1);
	if (al[1] < bl[1])
		return (-1);
	if (al[1] > bl[1])
		return (1);
	if (al[2] == bl[2])
		return (0);
	if (al[2] == 1)
		return (-1);
	return (1);
}


/*
 * Send aaRESTART or aaROUTEI to all gateways.
 */
sendall(sig)
{
	register struct anets *an;
	register struct mgates *mg, *lmg;
	register n;
	int trys, count, rcount;
	fd_set fds;
	struct sockaddr_in tsin;

	tsin = sin;
	log("sendall(%s)", sig == aaRESTART ? "aaRESTART" : "aaROUTEI");
	aas.magic = htonl(aaMagic); /* setup send packet */
	aas.type = sig;
	if (sig == aaROUTEI) {
		if(arouteslen > ROUTEPKTSIZ) {
			count = ROUTEPKTSIZ;
			aas.flags = (arouteslen+ROUTEPKTSIZ-1)/ROUTEPKTSIZ;
		} else {
			count = arouteslen;
			aas.flags = 0;
		}
		bcopy(aroutes, (char *)aas.stuff, count);
	} else {
		aas.flags = 0;
		count = 0;
	}
	aas.count = htons((u_short)count);
	count += aaconfMinSize;
	/*
	 * send to each kbox in the table.
	 */
	mg = mgates;
	lmg = &mgates[nmgates];
	for (an = &anets[0], n = 0 ; n < nanets ; an++, n++) {
		while (mg < lmg && mg->last < an)
			mg++;
		if ((an->flags & typeFlags) == arouteKbox &&
		    mg < lmg && an > mg->first && an <= mg->last)
			continue;
		if ((an->flags & arouteKbox) == 0 || (an->flags & arouteEtalk))
			continue;
		trys = 0;
		tsin.sin_addr.s_addr = an->iaddr;
		if (sendto(s, (caddr_t)&aas, count, 0,
		    (struct sockaddr *)&tsin, sizeof tsin) < 0) {
		    	log("sendto %s failed (sendall)",
			    inet_ntoa(tsin.sin_addr));
			continue;
		}
		/*
		 * receive until we get a good reply or timeout.
		 */
		for (;;) {
			FD_ZERO(&fds);
			FD_SET(s, &fds);
			tvwait.tv_sec = 2;  /* select waits 2 seconds */
			tvwait.tv_usec = 0;
			if (select(NFDBITS, &fds, (fd_set *)0, (fd_set *)0,
			    &tvwait) != 1) {
				/* timeout */
				if (++trys < 4) {
					(void)sendto(s, (caddr_t)&aas, count, 0,
					    (struct sockaddr *)&tsin,
					    sizeof tsin);
					continue;
				}
				log("no response from %s",
				    inet_ntoa(tsin.sin_addr));
				break;
			}
			fsinlen = sizeof fsin;
			rcount = recvfrom(s, (caddr_t)&aa, sizeof aa,
			    0, (struct sockaddr *)&fsin, &fsinlen);
			if (rcount < 0) {
				if (errno == EINTR)
					continue;
				log("recv failed");
				exit(1);
			}
			if (rcount < aaconfMinSize 
			    || ntohl(aa.magic) != aaMagic)
				continue;
			if (aa.ipaddr != an->iaddr) {
				sendreply();
				continue;
			}
			/* our request got thru! */
			sendreply();
			break;
		}
	}
	log("sendall(%s) complete", sig==aaRESTART ? "aaRESTART" : "aaROUTEI");
}

/*
 * check over configurations
 *
*/
checkconfigs()
{
  int  i;
  struct anets *an;
  struct conf *cp;
  int rangestart = 0;
  int rangeconflict = 0;

  for (an = anets, i = 0; i < nanets; i++, an++) {
    if (an->confsize >= sizeof(struct conf)) {
      cp = (struct conf *)an->conf;
      if (!cp->startddpWKSUnix) {
	cp->startddpWKSUnix = htons(defddpWKSUnix);
	if (rangestart) {
	  if (rangestart != cp->startddpWKSUnix)
	    rangeconflict = 1;
	} else rangestart = cp->startddpWKSUnix;
      } else {
	if (rangestart) {
	  if (rangestart != cp->startddpWKSUnix)
	    rangeconflict = 1;
	} else rangestart = cp->startddpWKSUnix;
      }
    }
  }
  rangestart = ntohs((u_short)rangestart);
  if (rangeconflict)
    log("Conflict in UDP WKS range start!");
  else if (rangestart != 200 && rangestart != defddpWKSUnix) {
    log("WARNING - UDP WKS range start %d is non-standard",
	ntohs((u_short)rangestart));
  }
}

/*
 * Read atalkatab database file.  Avoid rereading the file if the
 * write date hasnt changed since the last time we read it.
 */
readtab()
{
	struct stat sbuf1, sbuf;
	register char *sp, *cpp;
	register i;
	char st[128], *cp;
	register struct anets *an;
	struct mgates *mg;
	iaddr_t iaddr;
	short sh;
	long l;

	if (fp == 0) {
		if ((fp = fopen(atalkatab, "r")) == NULL) {
			log("can't open %s", atalkatab);
			exit(1);
		}
	}
	if (fstat(fileno(fp), &sbuf1) < 0)
	  goto reopen;
	if (stat(atalkatab, &sbuf) < 0)
	  goto reopen;
	if (sbuf1.st_dev == sbuf.st_dev && /* different... */
	    sbuf1.st_ino == sbuf.st_ino && /* ...file */
	    sbuf1.st_size == size && /* nec? */
	    sbuf.st_size == size &&
	    sbuf1.st_mtime == modtime && /* nec? */
	    sbuf.st_mtime == modtime)
	  return;
reopen:	
  	if (fp)
          (void)fclose(fp);
	if ((fp = fopen(atalkatab, "r")) == NULL) {
		log("can't open %s", atalkatab);
		exit(1);
	}
	(void)fstat(fileno(fp), &sbuf);
	log("(re)reading %s", atalkatab);
	modtime = sbuf.st_mtime;
	size = sbuf.st_size;
	nanets = 0;
	an = &anets[-1];
	mg = NULL;
	nmgates = 0;
	linenum = 0;
	cp = 0;

	/*
	 * read and parse each line in the file.
	 */
	for (;;) {
		if (fgets(line, sizeof line, fp) == NULL)
			break;	/* done */
		if ((i = strlen(line)))
			line[i-1] = 0;	/* remove trailing newline */
		linep = line;
		linenum++;
		if (line[0] == '#' || line[0] == 0)	/* skip comment lines */
			continue;
		if (line[0] == ' ' || line[0] == '\t') 	/* process conf info */
			goto confinfo;
		/*
		 * lines not beginning with white space 
		 * represent a new net #
		 */
		if (++nanets > NANETS) {
			log("'anets' table length exceeded");
			exit(1);
		}
		an++;
		cp = an->conf;			/* store following lines here */
		an->confsize = 0;
		getfield(st, sizeof st, 0);
		an->net = htons((u_short)getashort(st));
		getfield(st, sizeof st, 0);
		i = 0;
		/* parse flags */
		for (cpp = st ; *cpp ; cpp++) {
			if (isupper(*cpp))
				*cpp = tolower(*cpp);
			switch (*cpp) {
			case 'c':
				i |= arouteCore;  break;
			case 'k':
				i |= arouteKbox;  break;
			case 'h':
				i |= arouteHost;  break;
			case 'm':
				if (mg != NULL)
					mg->last = an - 1;
				if (an->net == 0) {
					nanets--;
					an--;
					mg = NULL;
					cp = 0;
					goto nextline;
				}
				if (nmgates >= NMGATES-1) {
					log("'mgates' table length exceeded");
					exit(1);
				}
				mg = &mgates[nmgates++];
				mg->first = an;
				i |= arouteKbox;
				break;
			case 'n':
				i |= arouteNet;  break;
 		        case 'e':
				i |= (arouteEtalk|arouteKbox); break;
			case 'a':
				i |= (arouteHost|arouteAsync); break;
			case '0': case '1': case '2': case '3':
				i |= (*cpp - '0');  break;
			default:
				log("bad flag (%c) in %s, linenum %d",
				    *cpp, st, linenum);
				break;
			}
		}
		an->flags = i;
		getfield(st, sizeof st, 0);
		an->iaddr = getiaddr(st);
		getfield(an->zone, sizeof(an->zone), 0);

	nextline:;
		continue;

		/*
		 * lines beginning with white space
		 * are configuration data for gateway.
		 */
	confinfo:;
		if (cp == 0) {
			log("misplaced configuration information, linenum %d",
			    linenum);
			continue;
		}
		for (;;) {	/* for each field in line */
			int len;

			getfield(st, sizeof st, 1);
			sp = st;
			if (*sp == 0)
				break;
			if (an->confsize >= sizeof an->conf) {
				log("Excess config info, linenum %d", linenum);
				break;
			}
			if (isupper(*sp))
				*sp = tolower(*sp);
			switch (*sp++) {
			case '%':
			  /* escape code, one of 'n' */
			  if (isupper(*sp))
			    *sp = tolower(*sp);
			  switch (*sp++) {
			    int flag;
			  case 'n':
			    /* decide which type of network based on */
			    /* the offset in the config structure */
			    if (mg != NULL)
			      iaddr = mg->first->iaddr;
			    else
			      iaddr = an->iaddr;
			    if (cp - an->conf ==
				(char *)&conf_proto.atnete-(char *)&conf_proto)
			      flag = arouteNet;
			    else if (cp - an->conf ==
				(char*)&conf_proto.anetet-(char*)&conf_proto)
			      flag = arouteKbox|arouteEtalk;
			    else if (cp - an->conf ==
				(char*)&conf_proto.atneta-(char*)&conf_proto) {
			      flag = arouteKbox;
			      iaddr = an->iaddr;
			    } else {
			      log("%%N at unexpected offset %d, line %d\n",
				  cp - an->conf, linenum);
			      flag = arouteKbox;
			    }

			    if ((sh = getanet(flag, iaddr)) == 0)
				log("%%N not set, line %d\n", linenum);
			    bcopy((caddr_t)&sh, cp, sizeof(short));
			    cp += sizeof(short);
			    an->confsize += sizeof(short);
			    break;

			  default:
			    log("bad field type %s, linenum %d", st, linenum);
			    break;
			  }
			  break;
			  
			case 'i':
				/* IP address name or number */
				iaddr = getiaddr(sp);
				bcopy((caddr_t)&iaddr, cp, sizeof iaddr);
				cp += sizeof iaddr;
				an->confsize += sizeof iaddr;
				break;

			case 'l':
				l = htonl((u_long)atoii(sp));
				bcopy((caddr_t)&l, cp, 4);
				cp += 4;
				an->confsize += 4;
				break;

			case 's':
				sh = htons((u_short)getashort(sp));
				bcopy((caddr_t)&sh, cp, 2);
				cp += 2;
				an->confsize += 2;
				break;
			
			case 'c':
				*cp = atoii(sp);
				cp++;
				an->confsize++;
				break;

			case '"':
				len = strlen(sp) - 1; /* drop trailing " */
				bcopy(sp, cp, len);
				cp += len;
				*cp++ = 0;
				len++;
				an->confsize += len;
				break;

			case '\'': /* pascal string */
				len = strlen(sp) - 1; /* drop trailing ' */
				*cp++ = len;
				bcopy(sp, cp, len);
				cp += len;
				len++; /* skip over length field */
				an->confsize += len;
				break;
			
			default:
				log("bad field type %s, linenum %d",
					st, linenum);
				break;
			}
			continue;	/* get next field in line */
		}
		/* get next line */
	}
	/* end of file */
	checkconfigs();
	arouteslen = buildart((struct arouteTuple *)aroutes, sizeof aroutes);
	azoneslen = buildzones(azones, sizeof azones);
	aproxylen = buildproxy((long *)aproxy, sizeof aproxy/sizeof(long));
}


/*
 * Get next field from 'line' buffer into 'str'.  'linep' is the 
 * pointer to current position.
 *
 * Fields are white space separated, except within quoted strings.
 * If 'quote' is true the quotes of such a string are retained, otherwise
 * they are stripped.  Quotes are included in strings by doubling them.
 */
getfield(str, len, quote)
	char *str;
{
	register char *lp = linep;
	register char *cp = str;

	while (*lp == ' ' || *lp == '\t')
		lp++;	/* skip spaces/tabs */
	if (*lp == 0 || *lp == '#') {
		*cp = 0;
		return;
	}
	len--;	/* save a spot for a null */

	if (*lp == '"' || *lp == '\'') {		/* quoted string */
		register term = *lp;

		if (quote) {
			*cp++ = term;
			len -= 2;	/* one for now, one for later */
		}
		lp++;
		while (*lp) {
			if (*lp == term) {
				if (lp[1] == term)
					lp++;
				else
					break;
			}
			*cp++ = *lp++;
			if (--len <= 0) {
				log("string truncated: %s, linenum %d",
				    str, linenum);
				if (quote)
					*cp++ = term;
				*cp = 0;
				linep = lp;
				return;
			}
		}
		if (!*lp)
			log("unterminated string: %s, linenum %d",
			    str, linenum);
		else {
			lp++;	/* skip the terminator */

			if (*lp && *lp != ' ' && *lp != '\t' && *lp != '#') {
				log("garbage after string: %s, linenum %d",
				    str, linenum);
				while (*lp && *lp != ' ' &&
				    *lp != '\t' && *lp != '#')
					lp++;
			}
		}
		if (quote)
			*cp++ = term;
	} else {
		while (*lp && *lp != ' ' && *lp != '\t' && *lp != '#') {
			*cp++ = *lp++;
			if (--len <= 0) {
				log("string truncated: %s, linenum %d",
				    str, linenum);
				break;
			}
		}
	}
	*cp = 0;
	linep = lp;
}


/*
 * Ascii to integer, with base check.
 */
atoii(s)
	register char *s;
{
	int v;
	char *c;

	if (isupper(*s))
		*s = tolower(*s);
	c = "%d";
	if (*s == '0') {
		c = "%o";
		if (*++s == 0)
			return (0);
	}
	if (*s == 'x' || *s == 'h') {
		c = "%x";
		s++;
	}
	if (sscanf(s, c, &v) != 1)
		log("bad numeric field %s, linenum %d", s, linenum);
	return (v);
}


/*
 * Get an internet address as a hostname or dot format string.
 */
getiaddr(st)
	register char *st;
{
	iaddr_t iaddr;
	register struct hostent *host;

	if (isdigit(*st)) {
		if ((iaddr = inet_addr(st)) == -1 || iaddr == 0)
			log("bad ipaddress %s, linenum %d", st, linenum);
	} else {
		if ((host = gethostbyname(st)) == 0)
			log("bad hostname %s, linenum %d", st, linenum);
		bcopy(host->h_addr, (caddr_t)&iaddr, sizeof iaddr);
	}
	return (iaddr);
}


/*
 * Get a short number or address.
 */
getashort(st)
	register char *st;
{
	register char *cp;

	if ((cp = index(st, '.')) == 0)
		return (atoii(st));
	*cp++ = 0;
	return ((atoii(st)<<8) | atoii(cp));
}


#ifdef VARARGS
#include <varargs.h>

/*
 * log an error message		(new version, for intelligent hosts)
 */
/* VARARGS */
log(va_alist)
	va_dcl
{
	register char *fmt;
	va_list args;
	FILE *fp;
	long time(), tloc;
	struct tm *tm, *localtime();

	if (debug)
	  fp = stderr;
	else if (scanonly)
	  fp = stdout;
	else if ((fp = fopen(atalkalog, "a+")) == NULL)
	  return;

	(void)time(&tloc);
	tm = localtime(&tloc);
	Fprintf(fp, "%d/%d %02.2d:%02.2d ", tm->tm_mon + 1, tm->tm_mday,
		tm->tm_hour, tm->tm_min);
	va_start(args);
	fmt = va_arg(args, char *);
	(void)vfprintf(fp, fmt, args);
	va_end(args);
	(void)putc('\n', fp);
	if (fp != stderr && fp != stdout)
		(void)fclose(fp);
}

#else

/*
 * log an error message  (ancient version for hosts without VARARGS)
 */
/* VARARGS 1 */
log(fmt, args)
	char *fmt;
{
	FILE *fp;
	long time(), tloc;
	struct tm *tm, *localtime();

	if (debug)
	  fp = stderr;
	else if (scanonly)
	  fp = stdout;
	else if ((fp = fopen(atalkalog, "a+")) == NULL)
	  return;

	(void)time(&tloc);
	tm = localtime(&tloc);
	Fprintf(fp, "%d/%d %02.2d:%02.2d ", tm->tm_mon + 1, tm->tm_mday,
		tm->tm_hour, tm->tm_min);
	_doprnt(fmt, &args, fp);
	(void)putc('\n', fp);
	if (fp != stderr && fp != stdout)
		(void)fclose(fp);
}
#endif

/*
 * getanet - scan through the table so far and find an appletalk network 
 * number. Called from the %n fields of the config structure
 *
 * flag field:
 *    arouteNet - match an IP address, scan for a net route that matches addr
 *    arouteKbox - scan for a route using this host
 *    arouteKbox|arouteETalk - scan for an Ethertalk net using this host.
 *
 * Limitations: Only one bridge can be configured with a given Ethertalk
 * network number this way. Network numbers referenced must be defined before
 * use.
 */
getanet(flags, ia)
     iaddr_t ia;
{
  register int i;
  struct anets *a;

  /* Scan table in reverse order - slightly more efficent */
  for(i = nanets-1; i >= 0; --i) {
    a = &anets[i];

    if ((flags == arouteNet &&
	 ((a->flags & (arouteNet | arouteHost)) != 0) &&
	 match_net(a->flags, ia, a->iaddr)) ||
	(flags == arouteKbox &&
	 (a->flags & (arouteKbox|arouteEtalk)) == arouteKbox &&
	 a->iaddr == ia) ||
	(flags == (arouteKbox|arouteEtalk) &&
	 (a->flags & flags) == flags &&
	 a->iaddr == ia))
      return(a->net);
  } 
  log("unable to match appletalk net (%x), line %d\n", flags, linenum);
  return(0);
}

/*
 * match_net - match a network number, using mask field stored in
 * flags.
 */
/* ARGSUSED */				/* or might be one day */
match_net(flags, a1, a2)
     unsigned flags;
     iaddr_t a1, a2;
{
  return((a1 & node_mask) == (a2 & node_mask));
}

char *
hostnameof(addr)
  u_long addr;
{
  struct hostent *host;
  struct in_addr ipaddr;

  ipaddr.s_addr = addr;
  host = gethostbyaddr((char *)&ipaddr.s_addr, sizeof(ipaddr.s_addr), AF_INET);
  if (!host)
    return((char *)inet_ntoa(ipaddr));
  return(host->h_name);
}

zipinput()
{
  u_short buf[1000];
  u_short *bp;
#if 0
  int icount;
#endif
  int len;
  char *p;
  char *zn;
  int zl;
  struct ZIP zipin;
  struct ZIP *z;

#if 0
  icount = ntohs(aa.count);		/* remember incoming count */
#endif
  bcopy((char *)aa.stuff, (char *)&zipin, sizeof(struct ZIP));
  if (zipin.command != zipQuery || zipin.count == 0)
    return (-1);
#if 0 /* count is a byte, max value 255 * 2 + header 2 bytes.. must fit */
  if (zipin.count > (sizeof(aa.stuff)-sizeof(struct ZIP))/sizeof(u_short))
    return (-1);
#endif
  /* get the list of networks */
  bcopy((char *)(aa.stuff+sizeof(struct ZIP)), (char *)buf,
      (int)(zipin.count * sizeof(u_short)));
  bp = buf;
  z = (struct ZIP *)aa.stuff;
  z->command = zipReply;
  z->count = 0;
  len = sizeof(struct ZIP);
  for (p = (char *)(z+1); len<512-3 && zipin.count; zipin.count--, bp++) {
    if ((zn = zoneof(*bp)) == NULL)
      continue;
    bcopy((char *)bp, p, sizeof(u_short)); /* copy in network */
    p += sizeof(u_short);	/* move along */
    zl = *p = strlen(zn);	/* get zone length */
    p++;			/* bounce */
    if (len + sizeof(u_short) + 1 + zl > 512)
	break;
    len += sizeof(u_short) + 1;	/* move along */
    z->count++;
    bcopy(zn, p, zl);		/* copy in zone name */
    p += zl;			/* bounce p ...*/
    len += zl;			/* ... and len */
  }
  return(len);
}
